/***********************************************************************
 predict-base.cc

 This code is part of the source to HapMotif package for analyzing
 conserved regions in haploid sequences of genetic variations.  See the
 README files for more information on the methods, use of the programs,
 and references.  This source code and any associated files are released
 under the terms of the MIT License, reproduced below.

 Copyright (c) 2003 Russell Schwartz, Department of Biological
 Sciences, Carnegie Mellon University.

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************/

/***********************************************************************
$Date: 2003/09/15 14:19:08 $
$Revision: 1.7 $
$Log: predict-base.cc,v $
Revision 1.7  2003/09/15 14:19:08  cvsuser
Made correction of known sites an option for prediction code.
Updated READMEs to reflect recent code changes.
Cleaned up and re-commented code in preparation for a new release.

Revision 1.6  2003/08/22 17:42:04  cvsuser
Removing corrections for possible sequencing errors.

Revision 1.5  2003/07/16 23:09:47  cvsuser
Corrected some minor errors in usage output of some executables.

Revision 1.4  2003/06/30 21:24:24  cvsuser
Adding some more files.

Revision 1.3  2003/06/23 17:59:28  cvsuser
Commiting a whole bunch of files because I have not updated them in a
while.

Revision 1.2  2003/05/30 21:06:06  cvsuser
Added code for doing missing site prediction and informative SNP
selection.

************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <list>
#include <vector>
#include <map>
#include <algorithm>
#include <math.h>
#include <iostream>
#include <fstream>
#include "haplotype.h"
#include "motif.h"
#include "motiflib.h"
#include "predict-base.h"

const double LOGZERO=-1e12;

/*************************************************************************
 Description: Predict the most probable ancestral values for a haplotype
 sequence with missing sites.
 Input: h, the haplotype; motifProbs, a map of motifs to their log
 probabilities; match, the log prior probability of a match of any
 site's measured value to its ancestral value;  mismatch, the log prior
 probability of a mismatch of any site's measured value to its ancestral
 value; correct, a boolean that is true if known sites inferred to be
 incorrect are replaced and false otherwise
 Output: h, updated so all sites contain the values predicted to be their
 most probable ancestral values
 Note: This code has been changed from earlier implementations to make it
 an option to replace known sites
**************************************************************************/
void predictSequence (Haplotype &h, map<Motif,double> &motifProbs,
		      double match, double mismatch, bool correct=false)
{
 for (int i=0;i<((int) h.sequence.size());i++)
  if (h.sequence[i]=='?' || correct)
  {
   double bestScore=LOGZERO;
   char bestBase=h.sequence[i];

   // Find the most probable motif touching site i
   for (map<Motif,double>::iterator mi=motifProbs.begin();
        mi!=motifProbs.end();mi++)
   {
    double score=mi->second;
    int start=mi->first.start, end=mi->first.end;
    string seq=mi->first.sequence;

    if (start>i || end<i)
     continue;

    for (int j=start;j<=end;j++)
     if (h.sequence[j]!='?')
     {
      if (h.sequence[j]==seq[j-start])
       score+=match;
      else
       score+=mismatch;
     }

    if (score>bestScore)
    {
     bestScore=score;
     bestBase=seq[i-start];
    }
   }

   h.sequence[i]=bestBase;
  }
}

/*************************************************************************
 Description: Predict the most probable ancestral values for a set of
 haplotype sequences some of which may have missing sites.
 Input: haplotypes, the haplotype sequences; motifProbs, a map of motifs
 to their log probabilities; mut, the prior probability that an oberved
 site has different allele value than its predicted ancestor; correct, a
 boolean that is true if known sites inferred to be incorrect are replaced
 and false otherwise
 Output: haplotypes, updated so all sites contain the values predicted to
 be their most probable ancestral values
 Note: This code has been changed from earlier implementations to make it
 an option to replace known sites
**************************************************************************/
void predictBases(vector<Haplotype> &haplotypes, map<Motif,double> &motifProbs,
		  double mut, bool correct=false)
{
 double match=(mut<1.0) ? log(1-mut) : LOGZERO;
 double mismatch=(mut>0.0) ? log(mut) : LOGZERO;

 for (int i=0;i<((int) haplotypes.size());i++) 
  predictSequence(haplotypes[i],motifProbs,match,mismatch,correct);
}

/*************************************************************************
 Description: Prints out usage information and exits the program.
 Input: argv, the array of input arguments to the executable
**************************************************************************/
void usage (char **argv)
{
 cerr << "Usage: " << argv[0] << " [-i <input file> ] [-s <motif file>] [-m <mutation probability>] [-c]\n";
 cerr << "-i <input file>: input sequence file name (default: stdin)\n";
 cerr << "-s <motif file>: input motif file name (default: stdin)\n";
 cerr << "-m <mutation probability>: prob of mutation (default: 0)\n";
 cerr << "-c : specifies correction of known sites inferred to be wrong\n";
 exit(6);
}

/*************************************************************************
 Description: Interprets command-line arguments.
 Input: argc and argv, the arguments to main
 Output: inputFile, the source of haplotypes with missing sites to be
 filled in or the empty string if standard input is to be used;
 motifFile, the source of motifs and probabilities or the empty string
 if standard input is to be used; mut, the prior probability of mutation;
 correct, a flag specifying whether known sites should be corrected if
 they are inferred to be errors
**************************************************************************/
void processCommandLine (int argc, char **argv, string &inputFile,
			 string &motifFile, double &mut, bool &correct)
{
 int i=1;

 // Set defaults
 inputFile="";
 motifFile="";
 mut=0.0;
 correct=false;
 // Read command line arguments
 while (i<argc)
 {
  string flag(argv[i++]);

  if (flag == "-i")
  {
   if (i>=argc)
     usage(argv);
   else
    inputFile=string(argv[i++]);
  }
  else if (flag == "-s")
  {
   if (i>=argc)
     usage(argv);
   else
    motifFile=string(argv[i++]);
  }
  else if (flag == "-m")
  {
   if (i>=argc)
     usage(argv);
   else
    mut=atof(argv[i++]);
  }
  else if (flag == "-c")
   correct=true;
  else
   usage(argv);
 }
 if (mut<0.0 || mut>1.0)
 {
  cerr << "Mutation parameter must be between 0 and 1\n";
  exit(11);
 }
}

int main (int argc, char **argv)
{
 string inputFile,motifFile;
 double mut;
 map<Motif,double> motifProbs;
 vector<Haplotype> haplotypes;
 bool correct;

 // Get arguments
 processCommandLine(argc,argv,inputFile,motifFile,mut,correct);

 // Read motif file
 if (motifFile=="")
 {
  readMotifs(stdin,motifProbs);
 }
 else
 {
  FILE *f=fopen(motifFile.c_str(),"r");

  if (f==NULL)
  {
   cerr << "Could not open " << motifFile << " for read.\n";
   exit(1);
  }
  readMotifs(f,motifProbs);
  fclose(f);
 }

 // Read haplotype file
 if (inputFile=="")
 {
  readHaplotypes(stdin,haplotypes);
 }
 else
 {
  FILE *f=fopen(inputFile.c_str(),"r");

  if (f==NULL)
  {
   cerr << "Could not open " << inputFile << " for read.\n";
   exit(1);
  }
  readHaplotypes(f,haplotypes);
  fclose(f);
 }

 // Predict missing/erroneous bases
 predictBases(haplotypes,motifProbs,mut,correct);

 // Output the revised sequences
 for (unsigned i=0;i<haplotypes.size();i++)
  cout << haplotypes[i].count << " " << haplotypes[i].sequence << "\n";

 return(0);
}
