/***********************************************************************
 predict-global.cc

 This code is part of the source to HapMotif package for analyzing
 conserved regions in haploid sequences of genetic variations.  See the
 README files for more information on the methods, use of the programs,
 and references.  This source code and any associated files are released
 under the terms of the MIT License, reproduced below.

 Copyright (c) 2003 Russell Schwartz, Department of Biological
 Sciences, Carnegie Mellon University.

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************/

/***********************************************************************
$Date: 2003/09/15 14:19:08 $
$Revision: 1.4 $
$Log: predict-global.cc,v $
Revision 1.4  2003/09/15 14:19:08  cvsuser
Made correction of known sites an option for prediction code.
Updated READMEs to reflect recent code changes.
Cleaned up and re-commented code in preparation for a new release.

Revision 1.3  2003/08/22 17:42:04  cvsuser
Removing corrections for possible sequencing errors.

Revision 1.2  2003/07/16 23:09:47  cvsuser
Corrected some minor errors in usage output of some executables.

Revision 1.1  2003/06/30 21:24:24  cvsuser
Adding some more files.

************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <list>
#include <vector>
#include <map>
#include <algorithm>
#include <math.h>
#include <iostream>
#include <fstream>
#include "haplotype.h"
#include "motif.h"
#include "motiflib.h"
#include "predict-global.h"

// A number that is supposed to be much smaller than the
// log of any number greater than zero.
const double LOGZERO=-1e12;

/*************************************************************************
 Description: Find the optimal interpretation of a haplotype in terms of
 motifs.
 Input: haps, the array of all haplotypes; h, the specific haplotype being
 analyzed; mutProb, the mutation probability; motifProbs, a map from
 candidate motif indicies to the log fraction of sequences containing that
 candidate motif; motifs, a map of all motifs listing which are defined;
 maxLength, the maximum allowed motif length; finder, a 2D array of lists,
 indexing motifs by length-1 and starting position
 Output: parse, an ordered list of the motifs making up the optimal parse
 of h
 Returns: the log probability of the optimal parse of h
**************************************************************************/
double parseHaplotype (Haplotype &h, double mutProb,
		       map<Motif,double> &motifProbs, map<Motif,bool> &motifs,
		       list<Motif> &parse, unsigned maxLength,
		       vector<vector<list<Motif> > > &finder)
{
 int i,j;
 int seqlen=h.sequence.size();
 vector<list<Motif> > bestList(seqlen);
 vector<double> bestScores(seqlen);
 vector<int> parents(seqlen);
 vector<Motif> bestMotifs(seqlen);

 // Pick the best starting site
 double opt=LOGZERO;
 Motif optseq;
 int optparent;
 list<Motif>::iterator indices;

 // Iterate over candidate start bases
 for (indices=finder[0][0].begin();
      indices!=finder[0][0].end();
      indices++)
 {
  if (!motifs[*indices])
   continue;

  // Start out with the probability of the given base
  double current=motifProbs[*indices];

  // Add in the score of this base for this sequence to account
  // for possible mutation
  current+=seqScore(h.sequence.substr(0,1),indices->sequence,mutProb);

  // If this is better than previous options, record it
  if (current>opt)
  {
   opt=current;
   optseq=*indices;
  }
 }
 bestScores[0]=opt;
 bestMotifs[0]=optseq;
 parents[0]=-1;

 // Do the dynamic programming iteration
 for (i=1;i<seqlen;i++)
 {
  opt=LOGZERO;
  optparent=-1;
  optseq=Motif();

  // Search over all motifs ending at i
  for (j=i;j>=0 && (unsigned) (i-j)<maxLength;j--)
  {
   for (indices=finder[i-j][j].begin();indices!=finder[i-j][j].end();indices++)
   {
    if (!motifs[*indices])
     continue;

    // Start out with the probability of the given motif
    double current=motifProbs[*indices];

    // Add in the score of this motif for this sequence to account
    // for possible mutation
    current+=seqScore(h.sequence.substr(j,i-j+1),indices->sequence,mutProb);

    // Add in the score of the best previous set of motifs
    if (j>0)
      current+=bestScores[j-1];

    // If this is better than previous options, record it
    if (current>opt)
    {
     opt=current;
     optparent=j-1;
     optseq=*indices;
    }
   }
  }

  // Record the best solution for this value of i
  bestScores[i]=opt;
  bestMotifs[i]=optseq;
  parents[i]=optparent;
 }

 // Backtrack to get the parse
 while (!parse.empty())
  parse.pop_front();
 for (i=seqlen-1;i>=0;i=parents[i])
  parse.push_front(bestMotifs[i]);

 // Return the log of the best probability
 return(bestScores[seqlen-1]);
}

/*************************************************************************
 Description: Predict the most probable ancestral values for a set of
 haplotype sequences some of which may have missing sites.
 Input: haplotypes, the haplotype sequences; motifProbs, a map of motifs
 to their log probabilities; mut, the prior probability that an oberved
 site has different allele value than its predicted ancestor; correct, a
 boolean that is true if known sites inferred to be incorrect are replaced
 and false otherwise
 Output: haplotypes, updated so all sites contain the values predicted to
 be their most probable ancestral values
 Note: This code has been changed from earlier implementations to make it
 an option to replace known sites
**************************************************************************/
void predictBases (vector<Haplotype> &haps, map<Motif,double> &motifProbs,
		   double mutProb, bool correct=false)
{
 vector<vector<list<Motif> > > finder;
 map<Motif,bool> motifs;
 list<Motif> parse;
 unsigned maxLength=0;

 // Prepare finder, motifs, and maxLength
 finder.resize(haps[0].sequence.size());
 for (unsigned i=0;i<finder.size();i++)
  finder[i].resize(finder.size());
 for (map<Motif,double>::iterator mi=motifProbs.begin();mi!=motifProbs.end();
      mi++)
 {
  unsigned offset, start;

  // Update finder
  start=mi->first.start;
  offset=mi->first.end-start;
  finder[offset][start].push_front(mi->first);

  // Update maxLength
  if (offset+1>maxLength)
   maxLength=offset+1;

  // Update motifs
  motifs[mi->first]=true;
 }

 // Fill in each haplotype
 for (unsigned i=0; i<haps.size(); i++)
 {
  string tmp;

  // Parse haplotype based on motifs
  parseHaplotype(haps[i],mutProb,motifProbs,motifs,parse,maxLength,finder);
  // Fill in data from parse
  for (list<Motif>::iterator pi=parse.begin(); pi!=parse.end(); pi++)
   tmp+=pi->sequence;
  for (unsigned j=0;j<tmp.size();j++)
    if (haps[i].sequence[j]=='?' || correct)
     haps[i].sequence[j]=tmp[j];
 }
}

/*************************************************************************
 Description: Prints out usage information and exits the program.
 Input: argv, the array of input arguments to the executable
**************************************************************************/
void usage (char **argv)
{
 cerr << "Usage: " << argv[0] << " [-i <input file> ] [-s <motif file>] [-m <mutation probability>]\n";
 cerr << "-i <input file>: input sequence file name (default: stdin)\n";
 cerr << "-s <motif file>: input motif file name (default: stdin)\n";
 cerr << "-m <mutation probability>: prob of mutation (default: 0)\n";
 cerr << "-c : specifies correction of known sites inferred to be wrong\n";
 exit(6);
}

/*************************************************************************
 Description: Interprets command-line arguments.
 Input: argc and argv, the arguments to main
 Output: inputFile, the source of haplotypes with missing sites to be
 filled in or the empty string if standard input is to be used;
 motifFile, the source of motifs and probabilities or the empty string
 if standard input is to be used; mut, the prior probability of mutation;
 correct, a flag specifying whether known sites should be corrected if
 they are inferred to be errors
**************************************************************************/
void processCommandLine (int argc, char **argv, string &inputFile,
			 string &motifFile, double &mut, bool &correct)
{
 int i=1;

 // Set defaults
 inputFile="";
 motifFile="";
 mut=0.0;
 correct=false;
 // Read command line arguments
 while (i<argc)
 {
  string flag(argv[i++]);

  if (flag == "-i")
  {
   if (i>=argc)
     usage(argv);
   else
    inputFile=string(argv[i++]);
  }
  else if (flag == "-s")
  {
   if (i>=argc)
     usage(argv);
   else
    motifFile=string(argv[i++]);
  }
  else if (flag == "-m")
  {
   if (i>=argc)
     usage(argv);
   else
    mut=atof(argv[i++]);
  }
  else if (flag == "-c")
   correct=true;
  else
   usage(argv);
 }
 if (mut<0.0 || mut>1.0)
 {
  cerr << "Mutation parameter must be between 0 and 1\n";
  exit(11);
 }
}

int main (int argc, char **argv)
{
 string inputFile,motifFile;
 double mut;
 map<Motif,double> motifProbs;
 vector<Haplotype> haplotypes;
 bool correct;

 // Get arguments
 processCommandLine(argc,argv,inputFile,motifFile,mut,correct);

 // Read motif file
 if (motifFile=="")
 {
  readMotifs(stdin,motifProbs);
 }
 else
 {
  FILE *f=fopen(motifFile.c_str(),"r");

  if (f==NULL)
  {
   cerr << "Could not open " << motifFile << " for read.\n";
   exit(1);
  }
  readMotifs(f,motifProbs);
  fclose(f);
 }

 // Read input haplotype file
 if (inputFile=="")
 {
  readHaplotypes(stdin,haplotypes);
 }
 else
 {
  FILE *f=fopen(inputFile.c_str(),"r");

  if (f==NULL)
  {
   cerr << "Could not open " << inputFile << " for read.\n";
   exit(1);
  }
  readHaplotypes(f,haplotypes);
  fclose(f);
 }

 // Predict missing/erroneous bases
 predictBases(haplotypes,motifProbs,mut,correct);

 // Output the answer
 for (unsigned i=0;i<haplotypes.size();i++)
  cout << haplotypes[i].count << " " << haplotypes[i].sequence << "\n";

 return(0);
}
